#include "Luna10ALS.h"
#include "time.h"
#include "stdio.h"


Luna10ALS::Luna10ALS (OBJHANDLE hObj, int fmodel): VESSEL2M (hObj, fmodel)
{
	// mesh
	VECTOR3 MeshOffset = _V(0, 0, 0);
	ClearMeshes();
	int MeshID = AddMesh ("Luna10ALS", &MeshOffset);
	SetMeshVisibilityMode (MeshID, MESHVIS_ALWAYS);

	// override some VESSEL2 defaults for crash effects
	VvCrash = 30; 
	VhCrash = 30;
	VMaxBlast = 200; 
	QMax = 45000;

	// animations
	bAnimate = false;
	bPlayClang = true;
	AnimationPercent = 0;
	Timer = 0;
	DefineAnimations();

	// attachment point
	CreateAttachment( TOPARENT, _V(0,0,0), _V(0,0,-1), _V(1,0,0), "ForKTDU", NOTLOOSE );
}


// General initialization of general params, called before config file
void Luna10ALS::clbkSetClassCaps (FILEHANDLE cfg)
{
	// Set geometry
	SetSize (0.5);

	// Set physics 
	SetEmptyMass (245);
	SetCrossSections (_V(0.5, 0.5, 0.5));
	SetLiftCoeffFunc (0);
	SetSurfaceFrictionCoeff(10,10);

	SetPMI (_V(1.0, 1.0, .5));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
	SetRotDrag (_V(0.5, 0.5, 0.5));
	SetCW (0.2, 0.2, 0.2, 0.2);

		// surface: set for the "jetisoned" case
	SetSurfaceFrictionCoeff (10.0, 10.0);
	SetCOG_elev (1);
	SetTouchdownPoints (_V(0, 4, -1), _V(-4, -4, -1), _V(4, -4, -1));
}	


// read scenario line
void Luna10ALS::ParseScenarioLineEx(char *line, void *status)
{
	// timer
	if (!_strnicmp (line, "TIMER", 5))
		sscanf (line+5, "%lf", &Timer);

	// antenna animation
	else if (!_strnicmp (line, "ANTENNA", 7))
		sscanf (line+7, "%lf", &AnimationPercent);

	// VESSEL2M and VESSEL2 stuff
	else 
		VESSEL2M::ParseScenarioLineEx (line, status);
}


// save scenario file
void Luna10ALS::clbkSaveState (FILEHANDLE scn)
{
	// do inherited VESSEL2M stuff before writing custom stuff
	VESSEL2M::clbkSaveState (scn);

	// timer
	if (Timer > 0)
		oapiWriteScenario_float (scn, "TIMER", Timer);

	// antenna animation, including possible intermediate state
	if (AnimationPercent == 1)
		oapiWriteScenario_int (scn, "ANTENNA", (int)AnimationPercent);
	else if (AnimationPercent > 0)
		oapiWriteScenario_float (scn, "ANTENNA", AnimationPercent);
}


void Luna10ALS::FinishConfiguration()
{
	bAnimate = (AnimationPercent > 0);     // move antenna to its final or intermediate position right away...
	bPlayClang = (AnimationPercent < 1);   // ...but don't clang if already in final position

	// do VESSEL2M stuff
	VESSEL2M::FinishConfiguration();
}


int Luna10ALS::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{

	// only process keydown events
	if (!down) 
		return 0; 

	// shift combinations are reserved
	if (KEYMOD_SHIFT (kstate)) 
		return 0; 

	// control combinations are reserved
	if (KEYMOD_CONTROL (kstate))
		return 0; 

	// animate antenna
	if (key == OAPI_KEY_K) 
	{
			if (AnimationPercent  == 0)
				bAnimate = true;
	}

	// toggle playing International
	if (key == OAPI_KEY_P) 
	{
		if (IsPlaying3(SoundLibID, SND_INTL))
			StopVesselWave3(SoundLibID, SND_INTL);

		// play only if antenna is out
		else if (AnimationPercent  == 1)
			PlayVesselWave3(SoundLibID, SND_INTL, LOOP);
	}

	return 0;
}


void Luna10ALS::clbkPostStep(double simt, double simdt, double mjd)
{
	Simdt = simdt;

	// animate antenna
	if (bAnimate)
		Animate();

	// if International is already playing, we have to maintain the call in every timestep
	// to comply with fading and Space Scilence
	if (IsPlaying3(SoundLibID, SND_INTL))
		PlayVesselWave3(SoundLibID, SND_INTL, LOOP);

	if (Timer > 0)  // this block runs only until timer is decremented below zero
	{
		// sound pyro if we started timer and antenna is not already deployed
		if ( (Timer > 4.95) && (AnimationPercent  != 1) )
			PlayVesselWave3(SoundLibID, SND_PYRO);

		// timer done?
		if ( (Timer-=Simdt) < 0 )
		{
			// did not animate antenna yet? do it now
			if (AnimationPercent  == 0)
				bAnimate = true;

			// antenna already out? start playing International.
			else
				PlayVesselWave3(SoundLibID, SND_INTL, LOOP);
		}
	}

	// common processing for all VESSEL2M-derived objects
	TimestepForVESSEL2M();
}



// Enforce camera dir switch here:
bool Luna10ALS::clbkLoadGenericCockpit ()
{
	oapiSetDefRCSDisplay(0);
	SetCameraOffset(_V(0.5, 0, 0.0));
	SetCameraDefaultDirection(_V(0, 0, -1), PI/2);
	return true;
}



void Luna10ALS::GetVesselMessage(VesselMessage VM)
{
	// antenna animation timer upon detaching
	if (VM == VM_TIMER)
	{
		Timer = 5.00;
		bFocus = true;
		PlayVesselWave3(SoundLibID, SND_PYRO);
	}

	// inherited VESSEL2M stuff
	else
		VESSEL2M::GetVesselMessage(VM);
}



	// ===== ANIMATIONS =====


void Luna10ALS::DefineAnimations()
{
	MGROUP_ROTATE *Antenna;
	static UINT grAntenna[1] = {0}; 

	Antenna = new MGROUP_ROTATE(0, grAntenna, 1, _V(0.338, -0.29, -0.438), _V(-0.5,  -0.5, 0), (float)(-125*RAD));

	Animation = CreateAnimation (0);
	AddAnimationComponent (Animation, 0, 1, Antenna);
}


void Luna10ALS::Animate()
{
	// just started? fire pyro sound.
	if (AnimationPercent == 0)
		PlayVesselWave3(SoundLibID, SND_PYRO);

	double SpeedOpen = 0.1 + AnimationPercent*5;// ends at 5.1
	double dPercent = Simdt * SpeedOpen;

	if (AnimationPercent < 1.0)
		AnimationPercent = min (1.0, AnimationPercent + dPercent);
	else
	{
		bAnimate = false;

		if (bPlayClang)
			PlayVesselWave3(SoundLibID, SND_CLANG);

		Timer = 1.5;
	}

	SetAnimation (Animation, AnimationPercent);
}




		// SOUND
void Luna10ALS::clbkPostCreation (void)
{
	SoundLibID = ConnectToOrbiterSoundDLL3(GetHandle());

	SoundOptionOnOff3(SoundLibID, PLAYMAINTHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYLANDINGANDGROUNDSOUND, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAIRSPEED, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAMBIANCEWHENLANDED, true);

	SoundOptionOnOff3(SoundLibID, PLAYCOUNTDOWNWHENTAKEOFF, false);
	SoundOptionOnOff3(SoundLibID, PLAYWHENATTITUDEMODECHANGE, false);
	SoundOptionOnOff3(SoundLibID, PLAYGPWS, false);
	SoundOptionOnOff3(SoundLibID, PLAYHOVERTHRUST, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKINGSOUND, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADARBIP, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKLANDCLEARANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINAIRCONDITIONING, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINRANDOMAMBIANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADIOATC, false);
	SoundOptionOnOff3(SoundLibID, DISPLAYTIMER, false);

	RequestLoadVesselWave3(SoundLibID, SND_INTL, "Sound\\Luna9\\Luna10International.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_PYRO, "Sound\\Luna9\\Pyro.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_CLANG, "Sound\\Luna9\\Clang.wav", BOTHVIEW_FADED_FAR);
}


// ==============================================================
// API interface
// ==============================================================

DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	srand((unsigned)time( NULL ));
	return new Luna10ALS (hvessel, flightmodel);
}

DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) 
		delete (Luna10ALS*)vessel;
}



