#include "Luna9AB.h"
#include "time.h"
#include "stdio.h"


Luna9AB::Luna9AB (OBJHANDLE hObj, int fmodel): VESSEL2M (hObj, fmodel)
{
	// mesh
	VECTOR3 MeshOffset = _V(0, 0, 0);
	ClearMeshes();
	int MeshID = AddMesh ("Luna9AB", &MeshOffset);
	SetMeshVisibilityMode (MeshID, MESHVIS_ALWAYS);

	// attachment point
	CreateAttachment( TOPARENT, _V(0,0,0), _V(0,0,-1), _V(1,0,0), "ForALS", NOTLOOSE );

	// animations and defaults
	DefineAnimations();
	bAnimate = false;
	CurrentAnimType = SPHERIC;
	SphericStatus = INITIAL;
	FlatStatus = INITIAL;
	CurrentAnimStatus = INITIAL;
	AnimationPercent = 0;
	AnimSpeed = INFLATION_SPEED;
	dZTouch = Z_ORIGINAL - Z_INFLATED;
}


// General initialization of general params, called before config file
void Luna9AB::clbkSetClassCaps (FILEHANDLE cfg)
{
	// geometry
	SetSize (1);

	// physics (mostly aerodynamics), non-changing
	SetEmptyMass (10);
	SetLiftCoeffFunc (0);
	SetPMI (_V(1.0, 1.0, .5));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
	SetRotDrag (_V(0.5, 0.5, 0.5));
	SetCW (0.2, 0.2, 0.2, 0.2);

		// surface
	SetSurfaceFrictionCoeff (1.0, 1.0);
	SetCOG_elev (1);
}	



// read scenario line
void Luna9AB::ParseScenarioLineEx(char *line, void *status)
{
	// geometry states; variables obfuscated : users should not edit these
	if (!strnicmp (line, "CFG1_STATUS", 11))
		SphericStatus = FINAL;
	else if (!strnicmp (line, "CFG2_STATUS", 11))
		FlatStatus = FINAL;

	// vars for any active animation 
	else if (!strnicmp (line, "AN_PERCENT", 10))
		sscanf (line+10, "%lf", &AnimationPercent);
	else if (!strnicmp (line, "AN_SPEED", 8))
		sscanf (line+8, "%lf", &AnimSpeed);
	else if (!strnicmp (line, "AN_DZ", 5))
		sscanf (line+5, "%lf", &dZTouch);
	else if (!strnicmp (line, "AN_STATUS", 9))
		sscanf (line+9, "%d", &CurrentAnimStatus);
	else if (!strnicmp (line, "AN_TYPE", 7))
		sscanf (line+7, "%d", &CurrentAnimType);

	// VESSEL2M and VESSEL2 stuff
	else 
		VESSEL2M::ParseScenarioLineEx(line, status);
}


// save scenario file
void Luna9AB::clbkSaveState (FILEHANDLE scn)
{
	// do inherited VESSEL2M stuff before writing custom stuff
	VESSEL2M::clbkSaveState (scn);

	// animating right now!
	if (bAnimate)
	{
		oapiWriteScenario_float (scn, "AN_PERCENT", AnimationPercent);
		oapiWriteScenario_float (scn, "AN_SPEED", AnimSpeed);
		oapiWriteScenario_float (scn, "AN_DZ", dZTouch);
		oapiWriteScenario_int (scn, "AN_STATUS", CurrentAnimStatus);
		oapiWriteScenario_int (scn, "AN_TYPE", CurrentAnimType);
	}

	// ... or not animating right now - write only complete
	else
	{
		if (SphericStatus == FINAL)
			oapiWriteScenario_int (scn, "CFG1_STATUS", SphericStatus);
		if (FlatStatus == FINAL)
			oapiWriteScenario_int (scn, "CFG2_STATUS", FlatStatus);
	}
}


// called after all vessel in scn are created, but before first timestep
void Luna9AB::FinishConfiguration()
{
	if (SphericStatus == FINAL)
		FastForwardAnimation(SPHERIC);

	if (FlatStatus == FINAL)
		FastForwardAnimation(FLAT);

	// unfinished animation, just like the new one, we only pteempted the AnimationPercent value
	bAnimate = ( (AnimationPercent > 0) && (AnimationPercent < 1) );
	if (bAnimate)
	{
		if (CurrentAnimType == SPHERIC)
		{
			FastForwardAnimation(SPHERIC, AnimationPercent);
			CurrentAnimation = SphericAnimation;
		}

		else
		{
			FastForwardAnimation(FLAT, AnimationPercent);
			CurrentAnimation = FlatAnimation;
		}
	}

	UpdatePhysics();

	// do VESSEL2M stuff
	VESSEL2M::FinishConfiguration();
}


void Luna9AB::UpdatePhysics()
{
	// flatened?
	if (FlatStatus == FINAL)
	{
		VvCrash = 200; 
		VhCrash = 200;
		VMaxBlast = 1000; 
		QMax = DRand(90000, 5000);

		SetCrossSections (_V(0.55, 0.55, 0.55));
		SetTouchdownPoints (_V( 1, 1,-Z_FLATENED), _V( 1, -1,-Z_FLATENED), _V(-1, 0,-Z_FLATENED));
	}

	// inflated?
	else if (SphericStatus == FINAL)
	{
		VvCrash = 50; 
		VhCrash = 50;
		VMaxBlast = 1000; 
		QMax = DRand(20000, 5000);

		SetCrossSections (_V(0.55, 0.55, 0.55));
		SetTouchdownPoints (_V( 1, 1,-Z_INFLATED), _V( 1, -1,-Z_INFLATED), _V(-1, 0,-Z_INFLATED));
	}

	// original? (not flatened and not inflated)
	else
	{
		VvCrash = 30; 
		VhCrash = 30;
		VMaxBlast = 1000; 
		QMax = DRand(30000, 5000);

		SetCrossSections (_V(0.05, 0.05, 0.5));
		SetTouchdownPoints (_V( 1, 1,-Z_ORIGINAL), _V( 1, -1,-Z_ORIGINAL), _V(-1, 0,-Z_ORIGINAL));
	}
}


void Luna9AB::clbkPostStep(double simt, double simdt, double mjd)
{
	// store timestep for animation
	Simdt = simdt;

	// step animation, if any
	if (bAnimate)
		Animate();

	// if we detached from ALS, start timer sequence.
	// These checks will run all the time after deflation - 
	// but there is nothing to do n Orbiter anyway at that time, 
	// so extra load won't hurt anything
	if (!bAttached)
	{
		// detached, inflated - start deflation
		if ( SphericStatus == FINAL)
			AnimationStart(SPHERIC);

		// detached, deflated - start flatening
		else if ( FlatStatus == INITIAL)
			AnimationStart(FLAT);
	}

	// VESSEL2M stuff
	TimestepForVESSEL2M();
}



void Luna9AB::GetVesselMessage(VesselMessage VM)
{
	if ( (VM == VM_INFLATE) && (SphericStatus == INITIAL) )
		AnimationStart(SPHERIC);

	// inherited VESSEL2M stuff
	else
		VESSEL2M::GetVesselMessage(VM);
}







	// ===== ANIMATIONS =====


void Luna9AB::DefineAnimations()
{
	static UINT grBag[1] = {0}; 

	MGROUP_SCALE *Spheric;
	Spheric = new MGROUP_SCALE(0, grBag, 1, _V( -2, 0, 0), _V(1.1, 1.1, 1.1));
	SphericAnimation = CreateAnimation (0);
	AddAnimationComponent (SphericAnimation, 0, 1, Spheric);

	MGROUP_SCALE *Flat;
	Flat = new MGROUP_SCALE(0, grBag, 1, _V( 0, 0, 0), _V(1, 1, 0.1));
	FlatAnimation = CreateAnimation (0);
	AddAnimationComponent (FlatAnimation, 0, 1, Flat);
}


void Luna9AB::FastForwardAnimation(anim_type AnimType, double AnimPercent)
{
	if (AnimType == SPHERIC)
		SetAnimation (SphericAnimation, AnimPercent);
	else
		SetAnimation (FlatAnimation, AnimPercent);

	if (AnimPercent == 1)
		AnimationEnd(AnimType, FINAL);
}


void Luna9AB::Animate()
{
	double dPercent = AnimSpeed * Simdt;

	if (CurrentAnimStatus == INITIAL) // opening
	{
		if (AnimationPercent < 1.0)
			AnimationPercent = min (1.0, AnimationPercent + dPercent);
		else
			AnimationEnd(CurrentAnimType, FINAL);

		SetAnimation (CurrentAnimation, AnimationPercent);
	} 

	if (CurrentAnimStatus == FINAL) // closing
	{
		if (AnimationPercent > 0.0)
			AnimationPercent = max (0.0, AnimationPercent - dPercent);
		else
			AnimationEnd(CurrentAnimType, INITIAL);

		SetAnimation (CurrentAnimation, AnimationPercent);
	} 

	// update ground contact points elevation
	double ZTouch = Z_ORIGINAL - AnimationPercent * dZTouch;
	SetTouchdownPoints (_V( 1, 1,-ZTouch), _V( 1, -1,-ZTouch), _V(-1, 0,-ZTouch));
}


void Luna9AB::AnimationStart(anim_type AnimType)
{
	// already animating something? ignore new animation request until done.
	if (bAnimate)
		return;

	// set up animation. these values depend on animation object...
	if (AnimType == SPHERIC)
	{
		CurrentAnimStatus = SphericStatus;
		CurrentAnimation = SphericAnimation;

		dZTouch = Z_ORIGINAL - Z_INFLATED;
	}
	else  // FLAT
	{
		CurrentAnimStatus = FlatStatus;
		CurrentAnimation = FlatAnimation;

		dZTouch = Z_ORIGINAL - Z_FLATENED;
	}

	// different speeds for inflate-delate
	if ( 
		 ( (AnimType == SPHERIC) && (CurrentAnimStatus == INITIAL) )
		 ||
		 ( (AnimType == FLAT) && (CurrentAnimStatus == FINAL) )
	   )
	   AnimSpeed = INFLATION_SPEED;
	else
	   AnimSpeed = DEFLATION_SPEED;

	// ...these don't 
	CurrentAnimType = AnimType;
	AnimationPercent = CurrentAnimStatus;

	// cross-condition for restricting dependences between animations
	if ( 
		 ( (AnimType == SPHERIC) && (FlatStatus == FINAL) )
		 ||
		 ( (AnimType == FLAT) && (SphericStatus == FINAL) )
	   )
	   return;

	// finally, allow animation
	bAnimate = true;
}


void Luna9AB::AnimationEnd(anim_type AnimType, anim_status NewStatus)
{
	bAnimate = false;

	if (AnimType == SPHERIC)
		SphericStatus = NewStatus;
	else
		FlatStatus = NewStatus;

	UpdatePhysics();
}




// ==============================================================
// API interface
// ==============================================================

DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	srand((unsigned)time( NULL ));
	return new Luna9AB (hvessel, flightmodel);
}

DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) 
		delete (Luna9AB*)vessel;
}



