#include "Luna9KTDU.h"
#include "time.h"
#include "stdio.h"


Luna9KTDU::Luna9KTDU (OBJHANDLE hObj, int fmodel): VESSEL2M (hObj, fmodel)
{
	// override some VESSEL2 defaults for crash effects
	VvCrash = 30; 
	VhCrash = 30;
	VMaxBlast = 400; 
	QMax = 40000;

	// mesh
	VECTOR3 MeshOffset = _V(0, 0, 0);
	ClearMeshes();
	int MeshID = AddMesh ("Luna9KTDU", &MeshOffset);
	SetMeshVisibilityMode (MeshID, MESHVIS_ALWAYS);

	// start with down-looking camera
	CameraDir = CAMERA_DOWN;

	// onscreen annotation message window
	double WW = 1280;
	double WH = 1024;

	MessagePort = oapiCreateAnnotation ( true, 0.5, _V(1, 1, 1) );
	oapiAnnotationSetPos (MessagePort, 0, 130/WH, 1, 1);
	msgStartTime = 0;
	bMessageDisplayed = false;

	// initialize vars
	bAutopilot = true;
	AltTurn = 0;
	MccState = MCC_NONE;
	AlsType = LANDER;
	bFocusAll = false;
	bSwitchFocus = true;
	MainThGroupCount = 7;  // start with all chambers included
	ChkDoType = CDT_DETACHED;
	MaxTimeAccel = MAX_TIME_ACCEL;
	WreckMeshID = MSH_WRECK_KTDU_V2M;

	// initialize runnig vars for landing autopilot
	Lap = 0;
	Pmain = 1;
	ProgramPoint[0] = 0;

	for (int i = 0; i < 99; i++)
	{
		LApLand[i] = 0;
		VApLand[i] = 0;
	}

	// initialize runnig vars for MCC autopilot
	YMoonDirLast = 0;
	ZVelDirLast = 0;

	// override VESSEL2M default, only here in KTDU
	bFocusOffByDefault = false;
	bFocus = true;
	
	// animation defaults
	bAnimate = false;
	AnimStatus = INITIAL;
	AnimationPercent = 0;
	DefineAnimations();
}


// General initialization of general params, called before config file
void Luna9KTDU::clbkSetClassCaps (FILEHANDLE cfg)
{
	SURFHANDLE ExhaustTexture = oapiRegisterExhaustTexture ("Luna9\\exhaust_crcs");

	// Set geometry
	SetSize (3);

	// Set physics 
	SetEmptyMass (150);
	SetPMI (_V(2, 2.25, 0.65));
	SetPitchMomentScale (0.001);
	SetBankMomentScale (0.0001);

	// aerodynamics
	SetCrossSections (_V(3.5, 3.5, 1.3));
	SetLiftCoeffFunc (0);
	SetRotDrag (_V(0.5, 0.5, 1));   // rotation resistance in atmosphere  
	SetCW (0.3, 0.6, 1, 1);			// wind resistance

	// Surface
	SetTouchdownPoints (_V(0, 0.5, 4), _V(-4, 0, -4), _V(4, -3, -4));
	SetSurfaceFrictionCoeff (10, 10);

		// Main engine group: set main engine plus 4 chambers of steer engine plus 2 chambers of roll steering
	MainTank = CreatePropellantResource(FUEL_MASS);

		// ROLL CHAMBERS

	MainEngine[0] = CreateThruster(_V( STEER_ROLL_X, -STEER_ROLL_Y, STEER_ROLL_Z), XMINUS, THRUST_STEER_ROLL, MainTank, ISP_MAIN, ISP_MAIN);
	MainEngine[1] = CreateThruster(_V(-STEER_ROLL_X-0.005, -STEER_ROLL_Y, STEER_ROLL_Z), XPLUS, THRUST_STEER_ROLL, MainTank, ISP_MAIN, ISP_MAIN);

	AddExhaust(MainEngine[0], STEER_ROLL_EXH_LENGTH, STEER_ROLL_EXH_WIDTH, 0., ExhaustTexture);
	AddExhaust(MainEngine[1], STEER_ROLL_EXH_LENGTH, STEER_ROLL_EXH_WIDTH, 0., ExhaustTexture);
	
	PARTICLESTREAMSPEC RollExhaustStream = { 0,			// flags, reserved
		0.012,			// srcsize, particle size at creation, m
		20.0,			// srcrate, average particle generation rate, Hz   
		5.0,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		0.001,			// lifetime, average particle lifetime [s]
		1.5,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0 };
	AddExhaustStream(MainEngine[0], &RollExhaustStream);
	AddExhaustStream(MainEngine[1], &RollExhaustStream);

		// STEER CHAMBERS

	MainEngine[2] = CreateThruster(_V( STEER_XY, STEER_XY, STEER_Z), _V(0,0,1), THRUST_STEER, MainTank, ISP_MAIN, ISP_MAIN);
	MainEngine[3] = CreateThruster(_V(-STEER_XY, STEER_XY, STEER_Z), _V(0,0,1), THRUST_STEER, MainTank, ISP_MAIN, ISP_MAIN);
	MainEngine[4] = CreateThruster(_V( STEER_XY, -STEER_XY, STEER_Z), _V(0,0,1), THRUST_STEER, MainTank, ISP_MAIN, ISP_MAIN);
	MainEngine[5] = CreateThruster(_V(-STEER_XY, -STEER_XY, STEER_Z), _V(0,0,1), THRUST_STEER, MainTank, ISP_MAIN, ISP_MAIN);

	AddExhaust(MainEngine[2], STEER_EXH_LENGTH, STEER_EXH_WIDTH, 0., ExhaustTexture);
	AddExhaust(MainEngine[3], STEER_EXH_LENGTH, STEER_EXH_WIDTH, 0., ExhaustTexture);
	AddExhaust(MainEngine[4], STEER_EXH_LENGTH, STEER_EXH_WIDTH, 0., ExhaustTexture);
	AddExhaust(MainEngine[5], STEER_EXH_LENGTH, STEER_EXH_WIDTH, 0., ExhaustTexture);

	PARTICLESTREAMSPEC SteerExhaustStream = { 0,			// flags, reserved
		0.05,			// srcsize, particle size at creation, m
		0.1,			// srcrate, average particle generation rate, Hz   
		20,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		0.03,			// lifetime, average particle lifetime [s]
		5,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0 };
	AddExhaustStream(MainEngine[2], &SteerExhaustStream);
	AddExhaustStream(MainEngine[3], &SteerExhaustStream);
	AddExhaustStream(MainEngine[4], &SteerExhaustStream);
	AddExhaustStream(MainEngine[5], &SteerExhaustStream);

		// MAIN CHAMBER
	// make last, so that it can be disabled (excluded from main group) later in flight

	MainEngine[6] = CreateThruster(_V(0, 0, MAIN_Z), _V(0,0,1), THRUST_MAIN, MainTank, ISP_MAIN, ISP_MAIN);

	AddExhaust(MainEngine[6], MAIN_EXH_LENGTH, MAIN_EXH_WIDTH, 0., ExhaustTexture);

	PARTICLESTREAMSPEC MainExhaustStream = { 0,			// flags, reserved
		0.27,			// srcsize, particle size at creation, m
		0.1,			// srcrate, average particle generation rate, Hz   
		20,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		0.07,			// lifetime, average particle lifetime [s]
		7,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0 };
	AddExhaustStream(MainEngine[6], &MainExhaustStream);


	// Make a big main engine group including steering and roll engines, and, sometimes, main engine.
	// Defer the actual thruster groop creation until we know better...
//	MainThGroup = CreateThrusterGroup(MainEngine, MainThGroupCount, THGROUP_MAIN);

	// RCS group combines both virtual RCS and gas RCS thrusters, only one of each will be used at specific times
	THRUSTER_HANDLE thRollLeft[2], thRollRight[2], thYawLeft[4], thYawRight[4], thPitchDown[3], thPitchUp[3];

		// virtual RCS group
	VirtualRCSTank = CreatePropellantResource(VIRT_RCS_FUEL_MASS);

	// ROLL
	thRollRight[0] = CreateThruster (_V(STEER_ROLL_X, -STEER_ROLL_Y, STEER_ROLL_Z), XMINUS, THRUST_VIRT_RSC_ROLL, VirtualRCSTank, ISP_VIRT_RCS, ISP_VIRT_RCS);
	AddExhaust (thRollRight[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

	thRollLeft[0] = CreateThruster (_V(-STEER_ROLL_X, -STEER_ROLL_Y, STEER_ROLL_Z), XPLUS, THRUST_VIRT_RSC_ROLL, VirtualRCSTank, ISP_VIRT_RCS, ISP_VIRT_RCS);
	AddExhaust (thRollLeft[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

	// YAW
	thYawLeft[0] = CreateThruster (_V(STEER_XY,  STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	thYawLeft[1] = CreateThruster (_V(STEER_XY, -STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	AddExhaust (thYawLeft[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);
	AddExhaust (thYawLeft[1], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

	thYawRight[0] = CreateThruster (_V(-STEER_XY,  STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	thYawRight[1] = CreateThruster (_V(-STEER_XY, -STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	AddExhaust (thYawRight[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);
	AddExhaust (thYawRight[1], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

	// PITCH
	thPitchDown[0] = CreateThruster (_V( STEER_XY, STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	thPitchDown[1] = CreateThruster (_V(-STEER_XY, STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	AddExhaust (thPitchDown[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);
	AddExhaust (thPitchDown[1], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

	thPitchUp[0] = CreateThruster (_V( STEER_XY, -STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	thPitchUp[1] = CreateThruster (_V(-STEER_XY, -STEER_XY, STEER_Z), ZPLUS, THRUST_VIRT_RSC, VirtualRCSTank, ISP_VIRT_RCS);
	AddExhaust (thPitchUp[0], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);
	AddExhaust (thPitchUp[1], VIRT_RCS_EXH_LENGTH, VIRT_RCS_EXH_WIDTH);

		// gas RCS group
	GasRCSTank = CreatePropellantResource(GAS_RCS_FUEL_MASS);

	// ROLL
	thRollLeft[1] = CreateThruster (_V( GAS_ROLL_X, -GAS_ROLL_Y, GAS_ROLL_Z), YPLUS, THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thRollLeft[1], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thRollLeft, 2, THGROUP_ATT_BANKLEFT);

	thRollRight[1] = CreateThruster (_V(-GAS_ROLL_X, -GAS_ROLL_Y, GAS_ROLL_Z), YPLUS, THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thRollRight[1], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thRollRight, 2, THGROUP_ATT_BANKRIGHT);

	// YAW
	thYawLeft[2] = CreateThruster (_V( GAS_YAW_X, -GAS_YAW_Y, GAS_YAW_Z_2), _V(-cos(70*RAD)*cos(50*RAD), cos(70*RAD)*cos(40*RAD), cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
	thYawLeft[3] = CreateThruster (_V(-GAS_YAW_X, -GAS_YAW_Y, GAS_YAW_Z), _V(cos(70*RAD)*cos(50*RAD), cos(70*RAD)*cos(40*RAD), -cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thYawLeft[2], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
//	AddExhaust (thYawLeft[3], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thYawLeft, 4, THGROUP_ATT_YAWLEFT);

	thYawRight[2] = CreateThruster (_V( GAS_YAW_X, -GAS_YAW_Y, GAS_YAW_Z), _V(-cos(70*RAD)*cos(50*RAD), cos(70*RAD)*cos(40*RAD), -cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
	thYawRight[3] = CreateThruster (_V(-GAS_YAW_X, -GAS_YAW_Y, GAS_YAW_Z_2),  _V(cos(70*RAD)*cos(50*RAD), cos(70*RAD)*cos(40*RAD), cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thYawRight[2], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
//	AddExhaust (thYawRight[3], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thYawRight, 4, THGROUP_ATT_YAWRIGHT);

	// PITCH
	thPitchDown[2] = CreateThruster (_V( GAS_PITCH_X, -GAS_PITCH_Y, GAS_PITCH_Z_2), _V(0, cos(70*RAD), -cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thPitchDown[2], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thPitchDown, 3, THGROUP_ATT_PITCHDOWN);

	thPitchUp[2] = CreateThruster (_V( GAS_PITCH_X, -GAS_PITCH_Y, GAS_PITCH_Z), _V(0, cos(70*RAD), cos(20*RAD)), THRUST_GAS_RSC, GasRCSTank, GAS_RCS_ISP);
//	AddExhaust (thPitchUp[2], GAS_RCS_EXH_LENGTH, GAS_RCS_EXH_WIDTH);
	CreateThrusterGroup (thPitchUp, 3, THGROUP_ATT_PITCHUP);

	// particles
	PARTICLESTREAMSPEC GasExhaustStream = { 0,			// flags, reserved
		0.005,			// srcsize, particle size at creation, m
		50.0,			// srcrate, average particle generation rate, Hz   
		1.0,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		0.001,			// lifetime, average particle lifetime [s]
		0.5,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0 };
	
	AddExhaustStream(thRollLeft[1], &GasExhaustStream);
	AddExhaustStream(thRollRight[1], &GasExhaustStream);
	AddExhaustStream(thYawLeft[2], &GasExhaustStream);
	AddExhaustStream(thYawLeft[3], &GasExhaustStream);
	AddExhaustStream(thYawRight[2], &GasExhaustStream);
	AddExhaustStream(thYawRight[3], &GasExhaustStream);
	AddExhaustStream(thPitchDown[2], &GasExhaustStream);
	AddExhaustStream(thPitchUp[2], &GasExhaustStream);
}	



// Read status from scenario file
void Luna9KTDU::ParseScenarioLineEx(char *line, void *status)
{
	char tempStr[256];

	// attachment names
	if (!strnicmp (line, "PO_ASTRO", 8))
		sscanf (line+8, "%s", &PoAstroName);
	else if (!strnicmp (line, "PO_RADIO", 8))
		sscanf (line+8, "%s", &PoRadioName);
	else if (!strnicmp (line, "ALS", 3))
		sscanf (line+3, "%s", &AlsName);

	// focus
	else if (!strnicmp (line, "FOCUS_ALL", 9))
		sscanf (line+9, "%d", &bFocusAll);
	else if (!strnicmp (line, "SWITCH_FOCUS", 12))
		sscanf (line+12, "%d", &bSwitchFocus);

	// mission: satellie  or lander?
	else if (!strnicmp (line, "MISSION", 7))
	{
		sscanf (line+7, "%s", &tempStr);
		if (tempStr[0] == 'S')
			AlsType = SATELLITE;
	}

	// camera, engine configuration
	else if (!strnicmp (line, "CAMERA", 6))
		sscanf (line+6, "%d", &CameraDir);
	else if (!strnicmp (line, "CFG_MAIN_E", 10))
		sscanf (line+10, "%d", &MainThGroupCount);

	// landing probe animation
	else if (!strnicmp (line, "CFG_STATE", 9))
		sscanf (line+9, "%d", &AnimStatus);
	else if (!strnicmp (line, "CFG_PROBE", 9))
		sscanf (line+9, "%lf", &AnimationPercent);

	// autopilots

	else if (!strnicmp (line, "AUTOPILOT", 9))
		sscanf (line+9, "%d", &bAutopilot);
	
	else if (!strnicmp (line, "CFG_CHECK_DO", 12))
		sscanf (line+12, "%d", &ChkDoType);

	else if (!strnicmp (line, "MCC_STATE", 9))
		sscanf (line+9, "%d", &MccState);

	else if (!strnicmp (line, "TIMER", 5))
		sscanf (line+5, "%lf", &Timer);

	else if (!strnicmp (line, "ALT_TURN", 8))
		sscanf (line+8, "%lf", &AltTurn);

	else if (!strnicmp (line, "ALT_BRAKE", 9))
		sscanf (line+9, "%lf", &AltBrake);

	else if (!strnicmp (line, "MAX_TIME_ACCEL", 14))
		sscanf (line+14, "%lf", &MaxTimeAccel);

	else if (!strnicmp (line, "INI_PERI", 8))
		sscanf (line+8, "%lf", &InitialPericenter);

	else if (!strnicmp (line, "PAST_PERI", 9))
		sscanf (line+9, "%lf", &PastStepPericenter);

	else if (!strnicmp (line, "LAP", 3))
		sscanf (line+3, "%lf", &Lap);


	// autopilot program
	else if (!strnicmp (line, "PROGRAM", 7)) 
	{
		double L, V;
		int i;
        sscanf (line+7, "%ld %lf %lf", &i, &V, &L);

		LApLand[i] = L; 
		VApLand[i] = V; 
	}

		// VESSEL2M and VESSEL2 stuff
	else 
		VESSEL2M::ParseScenarioLineEx(line, status);
}


// Save status to scenario file
void Luna9KTDU::clbkSaveState (FILEHANDLE scn)
{
	// do inherited VESSEL2M stuff before writing custom stuff
	VESSEL2M::clbkSaveState (scn);

	// attachment names
	if (PoAstroName[0] != NULL)
		oapiWriteScenario_string (scn, "PO_ASTRO", PoAstroName);
	if (PoRadioName[0] != NULL)
		oapiWriteScenario_string (scn, "PO_RADIO", PoRadioName);
	if (AlsName[0] != NULL)
		oapiWriteScenario_string (scn, "ALS", AlsName);

	// focus
	if (bFocusAll)
		oapiWriteScenario_int (scn, "FOCUS_ALL", bFocusAll);
	if (!bSwitchFocus)
		oapiWriteScenario_int (scn, "SWITCH_FOCUS", bSwitchFocus);

	// mission: satellie  or lander?
	if (AlsType == SATELLITE)
		oapiWriteScenario_string (scn, "MISSION", "SATELLITE");

	// camera, engine configuration
	if (CameraDir != CAMERA_DOWN)
		oapiWriteScenario_int (scn, "CAMERA", CameraDir);
	if (MainThGroupCount != 7)
		oapiWriteScenario_int (scn, "CFG_MAIN_E", MainThGroupCount);

	// landing probe animation
	if (AnimStatus != INITIAL)
		oapiWriteScenario_int (scn, "CFG_STATE", AnimStatus);
	if ( (AnimationPercent > 0 ) && (AnimationPercent < 1) )
		oapiWriteScenario_float (scn, "CFG_PROBE", AnimationPercent);

	// AUTOPILOTS

	// if autopilot is on, save current CheckDo state, otherwise save last-remembered state
	if (bAutopilot)
		SaveCheckDo();
	else
		oapiWriteScenario_int (scn, "AUTOPILOT", bAutopilot);

	if (ChkDoType != CDT_DETACHED)
		oapiWriteScenario_int (scn, "CFG_CHECK_DO", ChkDoType);

	// MCC 
	if (MccState != MCC_NONE)
	{
		oapiWriteScenario_int (scn, "MCC_STATE", MccState);
		oapiWriteScenario_float (scn, "INI_PERI", InitialPericenter);

		if (PastStepPericenter > 0)
			oapiWriteScenario_float (scn, "PAST_PERI", PastStepPericenter);
	}

	if (Timer > 0)
		oapiWriteScenario_float (scn, "TIMER", Timer);

	if (AltTurn > 0)
		oapiWriteScenario_float (scn, "ALT_TURN", AltTurn);

	if (AltBrake > 0)
		oapiWriteScenario_float (scn, "ALT_BRAKE", AltBrake);

	if (MaxTimeAccel != MAX_TIME_ACCEL)
		oapiWriteScenario_float (scn, "MAX_TIME_ACCEL", MaxTimeAccel);

	// autopilot program
	if ( VApLand[0] != 0)
	{
		oapiWriteScenario_float (scn, "LAP", Lap);

		for (int i = 0; i < 99; i++)
		{
			if (VApLand[i] > 0)
			{
				sprintf(ProgramPoint, "%d %.13f %.13f", i, VApLand[i], LApLand[i]);
				oapiWriteScenario_string(scn, "PROGRAM", ProgramPoint);
			}
		}
	}

}


// called after all vessel in scn are created, but before first timestep
void Luna9KTDU::FinishConfiguration()
{
	MainThGroup = CreateThrusterGroup(MainEngine, MainThGroupCount, THGROUP_MAIN);


	// create attachment points on us - always, or it screws many things
	hForPoAstro = CreateAttachment( TOCHILD, _V(0,0.739,0), _V(0,0,-1), _V(1,0,0), "ForPoAstro", NOTLOOSE );
	hForPoRadio = CreateAttachment( TOCHILD, _V(0,-0.739,0), _V(0,0,-1), _V(1,0,0), "ForPoRadio", NOTLOOSE );

	if (AlsType == LANDER)
		hForAls = CreateAttachment( TOCHILD, _V(0, 0, 1.24), _V(0,0,1), _V( cos(45*RAD),-cos(45*RAD), 0), "ForAls", NOTLOOSE );
	else
		hForAls = CreateAttachment( TOCHILD, _V(0, 0, 1.59), _V(0,0,1), _V( -1, 0, 0), "ForAls", NOTLOOSE );

	// finish vessel "assembly"
	SetupAls();
	SetupPoAstro();
	SetupPoRadio();

	// probe configuration
	if (AlsType == LANDER)
	{
		if (AnimStatus == EXTENDED)
			SetAnimation(ExtendAnimation, 1);
		else if (AnimStatus == BROKEN)
			Land();
		else if ( (AnimationPercent > 0) && (AnimationPercent < 1) )
			AnimationStart();
	}

	UpdatePhysics();

	// propagate focus enabling
	if (bFocusAll)
	{
		SendVesselMessage(PoAstroName, VM_CASCADE_FOCUS_ENABLE);
		SendVesselMessage(PoRadioName, VM_CASCADE_FOCUS_ENABLE);
		SendVesselMessage(AlsName, VM_CASCADE_FOCUS_ENABLE);
	}


	// Autopilots
	hMoon = oapiGetGbodyByName ("Moon");
	
	if (AlsType == SATELLITE)
		hEarth = oapiGetGbodyByName ("Earth"); 

	// if autopilot is off, we still remember the state in which i was turned off
	if (bAutopilot)
	{
		RestoreCheckDo();
	
		// restore retrograde, if applicable
		if ( (ChkDoType > CDT_TURN) && (ChkDoType != CDT_INERT) )
			ActivateNavmode(NAVMODE_RETROGRADE);
	
		// restore "one-time" autopilot messages
		ShowPermanentString(ChkDoType, IGNORE_FOCUS);
	}
	else
	{
		ChkDoType = CDT_INERT;
		pCheckCondition = &Luna9KTDU::CheckInert;
		pDoAction = NULL;
	}

	//VESSEL2M stuff
	VESSEL2M::FinishConfiguration();
}


int Luna9KTDU::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{

	// only process keydown events
	if (!down) 
		return 0; 

	// shift combinations are reserved
	if (KEYMOD_SHIFT (kstate)) 
		return 0; 

	// control combinations are usually reserved
	if (KEYMOD_CONTROL (kstate))
	{
		// toggle focus switch to ALS after its jetison
		if (key == OAPI_KEY_J) 
		{
			bSwitchFocus = !bSwitchFocus;

			if (bSwitchFocus)
				ShowAnnotationMessage("Switch focus to ALS after separation? YES!");
			else
				ShowAnnotationMessage("Switch focus to ALS after separation? NO!");
		}

		// propagate focus enabling for all components
		else if (key == OAPI_KEY_F) 
		{
			bFocusAll = true;
			
			SendVesselMessage(PoAstroName, VM_CASCADE_FOCUS_ENABLE);
			SendVesselMessage(PoRadioName, VM_CASCADE_FOCUS_ENABLE);
			SendVesselMessage(AlsName, VM_CASCADE_FOCUS_ENABLE);

			ShowAnnotationMessage("Focus is enabled for all spacecraft components.");
		}		

		return 0; 
	}

	// toggle internal view
	if (key == OAPI_KEY_C) 
	{

		if (CameraDir == CAMERA_DOWN)
			ToggleOnboardCamera(CAMERA_UP);

		else if (CameraDir == CAMERA_UP)
			ToggleOnboardCamera(CAMERA_DOWN);
		
		return 0;
	}

	// jetison components
	else if (key == OAPI_KEY_J) 
	{
		if (bAutopilot)
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");

		else if (GetAttachmentStatus(hForPoAstro))
			SeparatePoAstro();

		else if (GetAttachmentStatus(hForPoRadio))
			SeparatePoRadio();

		else if (GetAttachmentStatus(hForAls))
			SeparateAls();

		else
			SendVesselMessage(AlsName, VM_SEPARATE_CHILD);

		return 0; 
	}

	// inflate bags
	else if (key == OAPI_KEY_B) 
	{
		if (bAutopilot)
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");
		else
			InflateAirbags();
	}

	// enable-disable main chamber of the engine
	else if (key == OAPI_KEY_E) 
	{
		if (bAutopilot)
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");
		else
			ToggleMainThrustChamber();
	}

	// enable-disable autopilot
	else if (key == OAPI_KEY_M) 
		ToggleAutopilot();

	// extend landing probes
	else if (key == OAPI_KEY_K) 
	{
		if (bAutopilot)
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");
		else
			AnimationStart();
	}

	// manual debug
	else if (key == OAPI_KEY_D) 
		DebugSomething();

	return 0;
}


int Luna9KTDU::clbkConsumeDirectKey (char *keystate)
{
	if (bAutopilot)
	{
		// any of the following keys should be disregarded in autopilot mode
		if	(
				KEYDOWN (keystate, OAPI_KEY_ADD) + 
				KEYDOWN (keystate, OAPI_KEY_SUBTRACT) +
				KEYDOWN (keystate, OAPI_KEY_MULTIPLY) 
			)
		{
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");
			return 1;
		}
	}
	
	return 0;
}



// Enforce camera dir switch here:
bool Luna9KTDU::clbkLoadGenericCockpit ()
{
	oapiSetDefRCSDisplay(0);
	ToggleOnboardCamera(CameraDir);
	return true;
}



void Luna9KTDU::clbkFocusChanged (bool getfocus, OBJHANDLE hNewVessel, OBJHANDLE hOldVessel)
{
	// if we are loosing focus, clear the output
	if (hOldVessel == GetHandle())
		sprintf(oapiDebugString(), "");
}



void Luna9KTDU::clbkPostStep(double simt, double simdt, double mjd)
{
	// store timestep variables
	Simdt = simdt;
	Alt = GetAltitude();

	// generic condition/action pair, different at different times of the trajectory
	if ( (*this.*pCheckCondition)() )
		(*this.*pDoAction)();

	// step animation, if any
	if (bAnimate)
		Animate();

	// extended toucdown probes have a few special considerations
	if (AnimStatus == EXTENDED)
	{
		// give landing probe chance to work regardless of the autopilot state
		if ( ChkDoType != CDT_TOUCHDOWN )
		{
			if ( CheckTouchdown() )
				DoArrival();
		}

		// burn landing probes
		if ( (MainThGroupCount == 7) && (GetThrusterGroupLevel(MainThGroup) > 0) )
		{
			SetAnimation(ExtendAnimation, 0.50);
			SetAnimation(Brake1Animation, 1);
			SetAnimation(Brake2Animation, 1); 
	
			AnimStatus = BROKEN;

			ShowAnnotationMessage("TOUCHDOWN PROBES DESTROYED!");
		}
	}

	// virtual RCS should only work when the main engine group is runnng!
	// instead of messing with thruster groups, we simply play tricks with fuel supply (it is virtual anyway)
	if (GetThrusterGroupLevel(MainThGroup) == 0) 
		SetPropellantMass(VirtualRCSTank, 0);
	else
		SetPropellantMass(VirtualRCSTank, 1);

	// message display timer - always give it a chance to finish
	if (bMessageDisplayed)
	{
		if ( (oapiGetSysTime() - msgStartTime) > MSG_DISPLAY_TIME )
		{
			oapiAnnotationSetText (MessagePort, "");
			bMessageDisplayed = false;
		}
	}

	// common processing for all VESSEL2M-derived objects
	TimestepForVESSEL2M();
}



	// CHECK-DO PAIRS


// are we still attached to parent object (block L)?
bool Luna9KTDU::CheckDetach()
{
	if ( 
			(AttachmentCount(TOPARENT) > 0) && 
			(GetAttachmentStatus(GetAttachmentHandle(TOPARENT, 0)) == NULL) &&
			(Simdt > 0)
		)
		return true;

	else 
		return false;
}


// detached from block L!
void Luna9KTDU::DoDetach()
{
	bAttached = false;

	// next milestone: enter Moon influence spere
	pCheckCondition = &Luna9KTDU::CheckMoonInfluence;
	pDoAction = &Luna9KTDU::DoEnterMoonInfluence;

	PlayVesselWave3(SoundLibID, SND_PYRO);
}


// did we enter Moon influence spere?
bool Luna9KTDU::CheckMoonInfluence()
{
	return ( GetSurfaceRef() == hMoon );
}


// entered Moon influence spere!
void Luna9KTDU::DoEnterMoonInfluence()
{
	// this function does not do anything useful, but it is helpful to keep it in the sequence
	// to mark the trajectory milestone - and to make sure that we get close enough to the Moon
	// before the rest of the sequence is activated.
	// Of course, nothing currently prevents us from dropping back to Earth 
	// after wandering into the Moon shpere...

	// output
	ShowAnnotationMessage("Entered Moon gravity.");
	ShowPermanentString(CDT_MCCALT);

	// next milestone: midcource correction maneuver
	pCheckCondition = &Luna9KTDU::CheckMCCAlt;
	pDoAction = &Luna9KTDU::DoMCCAlt;
}


// did we reach midcource correction altitude?
bool Luna9KTDU::CheckMCCAlt()
{
	return (Alt < ALT_MCC_NOMINAL);
}


// start midcource correction maneuver!
void Luna9KTDU::DoMCCAlt()
{
	if (MccIsNeeded())
	{
		SetTimeAccelOff();
		ShowAnnotationMessage("Midcourse correction maneuver started.");

		MccState = MCC_ORIENT;
	}
	else
		MccState = MCC_NONE;

	MaxTimeAccel = 1;

	// next milestone: end of midcourse correction maneuver
	pCheckCondition = &Luna9KTDU::CheckMCCEnd;
	pDoAction = &Luna9KTDU::DoMCCEnd;
}


// do all steps of midcourse correction maneuver?
bool Luna9KTDU::CheckMCCEnd()
{
	LimitTimeAccel();

	if (MccState == MCC_ORIENT)
		MccAutopilotOrient();

	else if (MccState == MCC_STEADY)
		MccAutopilotWait();

	else if (MccState == MCC_BURN)
		MccAutopilotBurn();

	return (MccState == MCC_NONE);
}


// midcourse correction maneuver is completed!
void Luna9KTDU::DoMCCEnd()
{
	AltTurn = ALT_TURN;
	MaxTimeAccel = MAX_TIME_ACCEL;


	// This might be a temporary condition for elliptic approach. 
	// We might not need it when we go hyperbolic... but we might also leave it as is.
	if (AlsType == SATELLITE)
	{
		// Get current Moon-based pericenter. GetPeriDist is still using Earth, so we use GetElements.
		ELEMENTS el;
		ORBITPARAM prm;
		GetElements (hMoon, el, &prm); 

		AltTurn = prm.PeD;	 // get pericenter alt
		AltTurn -= 1738000;  // deduct Moon radius
		AltTurn *= 2;        // target for double the pericenter alt to do a realistic (non-peri) burn
	}

	// output
	ShowAnnotationMessage("Midcourse correction complete.");
	ShowPermanentString(CDT_TIME1000);

	// next milestone: limited time acceleration
	pCheckCondition = &Luna9KTDU::CheckTimeAccel1000;
	pDoAction = &Luna9KTDU::DoTimeAccel1000;
}



// did we reach altitude with limited time acceleration?
bool Luna9KTDU::CheckTimeAccel1000()		
{
	return (Alt < ALT_TIMEACCEL_1000);
}

// update max time acceleration!
void Luna9KTDU::DoTimeAccel1000()		
{
	MaxTimeAccel = 1000;

	// for luna10, jump right through 100 time accel
	if (AlsType == SATELLITE)
		DoTimeAccel100();

	else // lander
	{
		// next milestone: limited time acceleration
		pCheckCondition = &Luna9KTDU::CheckTimeAccel100;
		pDoAction = &Luna9KTDU::DoTimeAccel100;
	}
}

// did we reach altitude with limited time acceleration?
bool Luna9KTDU::CheckTimeAccel100()
{
	LimitTimeAccel();

	return (Alt < ALT_TIMEACCEL_100);
}

// update max time acceleration!
void Luna9KTDU::DoTimeAccel100()		
{
	MaxTimeAccel = 100;

	// next milestone: retrograde turn altitude
	pCheckCondition = &Luna9KTDU::CheckTurn;
	pDoAction = &Luna9KTDU::DoTurn;
}



// did we reach the retrograde turn altitude?
bool Luna9KTDU::CheckTurn()
{
	LimitTimeAccel();

	/*	if (AlsType == LANDER)
		return (Alt < ALT_TURN);

	else // Luna-10
	{
		// %%% reserve the following for a hyprtbolic flyby
		// detect when there is a 90 degrees angle between Earth, Moon and us.
		VECTOR3 EarthPos, UsPos;
		oapiGetRelativePos (hEarth, hMoon, &EarthPos);
		oapiGetRelativePos (GetHandle(), hMoon, &UsPos);

		// angle between vectors
		double AngleEarthUsMoon = asin( dotp(EarthPos, UsPos) / (length(EarthPos) * length(UsPos)) ) * DEG;
		
		if (oapiGetFocusObject() == GetHandle())
			sprintf(oapiDebugString(), "checkTURN - angle %f", AngleEarthUsMoon);

		return (AngleEarthUsMoon > 90);
	}*/

	return (Alt < AltTurn);
}


// start retrograde turn!
void Luna9KTDU::DoTurn()
{
	SetTimeAccelOff();
	ActivateNavmode(NAVMODE_RETROGRADE);
	MaxTimeAccel = 10;
	AltTurn = 0;  // reset, so that it does not get saved in scn when not needed

	// caclulate breaking altitude or time
	if (AlsType == LANDER)
	{
		CalcAltBrake();

		// output
		ShowAnnotationMessage("Turn started.");
		ShowPermanentString(CDT_IGNALT);

		// next milestone: ignition altitude
		pCheckCondition = &Luna9KTDU::CheckIgnitionAlt;
		pDoAction = &Luna9KTDU::DoIgnitionAlt;
	}

	else // Luna10
	{
		ShowAnnotationMessage("Orbit insertion sequence started.");
		Timer = 60; // %%% will 1 minutes be always enough to complete turn?

		// next milestone: ignition time
		pCheckCondition = &Luna9KTDU::CheckIgnitionTime;
		pDoAction = &Luna9KTDU::DoIgnitionTime;
	}
}


// did we reach ignition altitude? ONLY LUNA-9!
bool Luna9KTDU::CheckIgnitionAlt()
{
	LimitTimeAccel();

	return (Alt < AltBrake);
}


// reached ignition altitude! ONLY LUNA-9!
void Luna9KTDU::DoIgnitionAlt()
{
	SetThrusterGroupLevel(MainThGroup, 1);
	MaxTimeAccel = 1;
	AltBrake = 0;

	// do we have radio bay (with air tanks) to inflate the airbags?
	if (GetAttachmentStatus(hForPoAstro))
	{
		InflateAirbags();
		Timer = 1; 

		// sleanup string, nothing more to report...
		if (oapiGetFocusObject() == GetHandle())
			sprintf(oapiDebugString(), "");

		// next milestone: parachuting
		pCheckCondition = &Luna9KTDU::CheckInflationEnd;
		pDoAction = &Luna9KTDU::DoInflationEnd;
	}
	else
	{
		if (oapiGetFocusObject() == GetHandle())
			ShowAnnotationMessage("EQUIPMENT BAYS DROPPED PREMATURELY, AIRBAGS NOT IFLATED!");

		// next milestone: skip inflation, jump to parachuting
		pCheckCondition = &Luna9KTDU::CheckParachuting;
		pDoAction = &Luna9KTDU::DoLowStage;
	}
}


// did we reach ignition time? ONLY LUNA-10!
bool Luna9KTDU::CheckIgnitionTime()
{
	LimitTimeAccel();

	if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "Ignition in %.0f s...", Timer);

	return ((Timer -= Simdt) < 0);
}


// reached ignition time! ONLY LUNA-10!
void Luna9KTDU::DoIgnitionTime()
{
	MaxTimeAccel = 1;
	SetThrusterGroupLevel(MainThGroup, 1);

	if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "");

	// next milestone: pericenter
	pCheckCondition = &Luna9KTDU::CheckPericenter;
	pDoAction = &Luna9KTDU::DoLowStage;
}


// is airbags inflation complete?
bool Luna9KTDU::CheckInflationEnd()
{
	LimitTimeAccel();
	
	LandingAutopilot();

	return (Timer -= Simdt) < 0;
}


// airbags inflation is complete!
void Luna9KTDU::DoInflationEnd()
{
	SeparatePoAstro();
	SeparatePoRadio();

	// next milestone: parachuting
	pCheckCondition = &Luna9KTDU::CheckParachuting;
	pDoAction = &Luna9KTDU::DoLowStage;
}


// did we reach parachuting velocity? ONLY LUNA-9!
bool Luna9KTDU::CheckParachuting()
{
	LimitTimeAccel();

	LandingAutopilot();

	VECTOR3 VHor;
	GetHorizonAirspeedVector(VHor);

	return (length(VHor) < 40);
}


// did we reach the minimal pericenter orbit? ONLY LUNA-10!
bool Luna9KTDU::CheckPericenter()
{
	LimitTimeAccel();

	double Pericenter;
	GetPeDist(Pericenter);

	return ( (Pericenter < 2088000) || (GetPropellantMass (MainTank) == 0) );   // 350 km or no fuel
}



// switch engine to low stage
void Luna9KTDU::DoLowStage()
{
	DisableMainThrustChamber();
	
	if (AlsType == LANDER)
	{
		AnimationStart(); // extend probe

		// next milestone: touchdown
		pCheckCondition = &Luna9KTDU::CheckTouchdown;
		pDoAction = &Luna9KTDU::DoArrival;
	}

	else // Luna10
	{
		Timer = 5;

		// next milestone: MECO
		pCheckCondition = &Luna9KTDU::CheckMECO;
		pDoAction = &Luna9KTDU::DoArrival;
	}
}


// did the probe end touch the ground? ONLY LUNA-9!
bool Luna9KTDU::CheckTouchdown()
{
	LimitTimeAccel();

	// check for the possible need to premuaturely shotdown the parachuting engine
	// going down too slowly (-1 m/s), risking to start accending again
	VECTOR3 VHor;
	GetHorizonAirspeedVector(VHor);
	if (VHor.y > -1)
		SetThrusterGroupLevel(MainThGroup, 0);

	if (AnimStatus == EXTENDED)
	{
		// effective height of the probe
		VECTOR3 Z1H;
		HorizonRot(ZPLUS, Z1H);
		double HProbe = 5 * Z1H.y;

		// probe end altitude
		double AltProbeEnd = Alt - HProbe;

		// touchdown?
		return (AltProbeEnd < 0);
	}
	else
		return false;
}


// did we reach MECO time? ONLY LUNA-10!
bool Luna9KTDU::CheckMECO()
{
	LimitTimeAccel();

	if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "MECO in %.0f s...", Timer);

	return ((Timer -= Simdt) < 0);
}


// arrived at destination!
void Luna9KTDU::DoArrival()
{
	SeparateAls();

	// permanently safe main engine
	SetThrusterGroupLevel(MainThGroup, 0);

	// clean up the message string
	if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "");

	// reset, so autopilot program won't be stored in scenario
	VApLand[0] = 0;
	MaxTimeAccel = MAX_TIME_ACCEL;

	// no next milestone: safe inert state
	pCheckCondition = &Luna9KTDU::CheckInert;
	pDoAction = NULL;
}


// don't really check anything, always return false
bool Luna9KTDU::CheckInert()
{
	return false;
}


void Luna9KTDU::SaveCheckDo()
{
	if (pCheckCondition == &Luna9KTDU::CheckDetach)
		ChkDoType = CDT_DETACHED;
	else if (pCheckCondition == &Luna9KTDU::CheckMoonInfluence)
		ChkDoType = CDT_MOONINFL;
	else if (pCheckCondition == &Luna9KTDU::CheckMCCAlt)
		ChkDoType = CDT_MCCALT;
	else if (pCheckCondition == &Luna9KTDU::CheckMCCEnd)
		ChkDoType = CDT_MCCEND;
	else if (pCheckCondition == &Luna9KTDU::CheckTimeAccel1000)
		ChkDoType = CDT_TIME1000;
	else if (pCheckCondition == &Luna9KTDU::CheckTimeAccel100)
		ChkDoType = CDT_TIME100;
	else if (pCheckCondition == &Luna9KTDU::CheckTurn)
		ChkDoType = CDT_TURN;
	else if (pCheckCondition == &Luna9KTDU::CheckIgnitionAlt)
		ChkDoType = CDT_IGNALT;
	else if (pCheckCondition == &Luna9KTDU::CheckIgnitionTime)
		ChkDoType = CDT_IGNTIME;
	else if (pCheckCondition == &Luna9KTDU::CheckInflationEnd)
		ChkDoType = CDT_INFLEND;
	else if (pCheckCondition == &Luna9KTDU::CheckParachuting)
		ChkDoType = CDT_PARA;
	else if (pCheckCondition == &Luna9KTDU::CheckPericenter)
		ChkDoType = CDT_PERI;
	else if (pCheckCondition == &Luna9KTDU::CheckTouchdown)
		ChkDoType = CDT_TOUCHDOWN;
	else if (pCheckCondition == &Luna9KTDU::CheckMECO)
		ChkDoType = CDT_MECO;
	else if (pCheckCondition == &Luna9KTDU::CheckInert)
		ChkDoType = CDT_INERT;
}


void Luna9KTDU::RestoreCheckDo()
{

	if (ChkDoType == CDT_DETACHED)
	{
		pCheckCondition = &Luna9KTDU::CheckDetach;
		pDoAction = &Luna9KTDU::DoDetach;
	}
	else if (ChkDoType == CDT_MOONINFL)
	{
		pCheckCondition = &Luna9KTDU::CheckMoonInfluence;
		pDoAction = &Luna9KTDU::DoEnterMoonInfluence;
	}
	else if (ChkDoType == CDT_MCCALT)
	{
		pCheckCondition = &Luna9KTDU::CheckMCCAlt;
		pDoAction = &Luna9KTDU::DoMCCAlt;
	}
	else if (ChkDoType == CDT_MCCEND)
	{
		pCheckCondition = &Luna9KTDU::CheckMCCEnd;
		pDoAction = &Luna9KTDU::DoMCCEnd;
	}
	else if (ChkDoType == CDT_TIME1000)
	{
		pCheckCondition = &Luna9KTDU::CheckTimeAccel1000;
		pDoAction = &Luna9KTDU::DoTimeAccel1000;
	}
	else if (ChkDoType == CDT_TIME100)
	{
		pCheckCondition = &Luna9KTDU::CheckTimeAccel100;
		pDoAction = &Luna9KTDU::DoTimeAccel100;
	}
	else if (ChkDoType == CDT_TURN)
	{
		pCheckCondition = &Luna9KTDU::CheckTurn;
		pDoAction = &Luna9KTDU::DoTurn;
	}
	else if (ChkDoType == CDT_IGNALT)
	{
		pCheckCondition = &Luna9KTDU::CheckIgnitionAlt;
		pDoAction = &Luna9KTDU::DoIgnitionAlt;
	}
	else if (ChkDoType == CDT_IGNTIME)
	{
		pCheckCondition = &Luna9KTDU::CheckIgnitionTime;
		pDoAction = &Luna9KTDU::DoIgnitionTime;
	}
	else if (ChkDoType == CDT_INFLEND)
	{
		pCheckCondition = &Luna9KTDU::CheckInflationEnd;
		pDoAction = &Luna9KTDU::DoInflationEnd;
	}
	else if (ChkDoType == CDT_PARA)
	{
		pCheckCondition = &Luna9KTDU::CheckParachuting;
		pDoAction = &Luna9KTDU::DoLowStage;
	}
	else if (ChkDoType == CDT_PERI)
	{
		pCheckCondition = &Luna9KTDU::CheckPericenter;
		pDoAction = &Luna9KTDU::DoLowStage;
	}
	else if (ChkDoType == CDT_TOUCHDOWN)
	{
		pCheckCondition = &Luna9KTDU::CheckTouchdown;
		pDoAction = &Luna9KTDU::DoArrival;
	}
	else if (ChkDoType == CDT_MECO)
	{
		pCheckCondition = &Luna9KTDU::CheckMECO;
		pDoAction = &Luna9KTDU::DoArrival;
	}
	else if (ChkDoType == CDT_INERT)
	{
		pCheckCondition = &Luna9KTDU::CheckInert;
		pDoAction = NULL;
	}
}





	// AUTOPILOTS

void Luna9KTDU::ToggleAutopilot()
{
	if (bAutopilot)
		DisableAutopilot();
	else
		EnableAutopilot();
}


void Luna9KTDU::DisableAutopilot()
{
	// save autopilot state
	SaveCheckDo();

	// disable autopilot state
	pCheckCondition = &Luna9KTDU::CheckInert;
	pDoAction = NULL;

	bAutopilot = false;

	// disable any possible residual navmodes
	DeactivateNavmode(NAVMODE_KILLROT);
	DeactivateNavmode(NAVMODE_HLEVEL);
	DeactivateNavmode(NAVMODE_PROGRADE);
	DeactivateNavmode(NAVMODE_RETROGRADE);
	DeactivateNavmode(NAVMODE_NORMAL);
	DeactivateNavmode(NAVMODE_ANTINORMAL);
	DeactivateNavmode(NAVMODE_HOLDALT);

	SetThrusterGroupLevel(THGROUP_MAIN, 0);
	SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);

	// show message
	ShowAnnotationMessage("Autopilot disabled!");

	// clear any ongoing autopilot output
	sprintf(oapiDebugString(), "");
}


void Luna9KTDU::EnableAutopilot()
{
	if (!CanReenableAutopilot())
		return;

	// restore autopilot state
	bAutopilot = true;

	RestoreCheckDo();

	if (ChkDoType >= CDT_TURN)
		ActivateNavmode(NAVMODE_RETROGRADE);

	// show message (most likely will be immediately overridden by autopilot)
	ShowAnnotationMessage("Autopilot enabled!");
}


bool Luna9KTDU::CanReenableAutopilot()
{
	if (AlsType == LANDER)
	{
		if ( (Alt < 50000) ||
			 ( (Alt < 200000) && (ChkDoType < CDT_TURN) )
			)
		{
			ShowAnnotationMessage("Autopilot cannot be re-enabled: too low!");
			return false;
		}
	}

	else if (AlsType == SATELLITE)
	{
		if (Alt < 3000000)
		{
			ShowAnnotationMessage("Autopilot cannot be re-enabled: too late!");
			return false;
		}
	}
			
	else if (GetPropellantMass(MainTank)/GetPropellantMaxMass(MainTank) <= 0.89)
	{
		ShowAnnotationMessage("Autopilot cannot be re-enabled: low fuel!");
		return false;
	}

	return true;
}

bool Luna9KTDU::MccIsNeeded()
{
	// orbit relative to moon
	ELEMENTS ElToMoon;
	ORBITPARAM OPtoMoon;
	GetElements(hMoon, ElToMoon, &OPtoMoon, 0, FRAME_EQU);

	// initialize InitialPericenter
	InitialPericenter = OPtoMoon.PeD;

	// initialize PastStepPericenter
	if ( (AlsType == SATELLITE) && (InitialPericenter < MCC_10_PERI_MIN) )
		PastStepPericenter = 0; // some really small number...
	else
		PastStepPericenter = 1000000000000; // some really big number...

		// is it too late? (maybe, if we zoomed through MCC window in manual mode...) 
	if (Alt < ALT_MCC_MAX)
		return false;

	// do we have enough fuel? 
	if (GetPropellantMass(MainTank)/GetPropellantMaxMass(MainTank) <= 0.95)
		return false;

	// Luna9: MCC is needed when pericenter radius (not altitude) is > 10 km
	else if (AlsType == LANDER)
		return (InitialPericenter > MCC_9_PERI_MIN);  // 10 km

	// Luna10: MCC is needed when pericenter altitude (not radius) is not btweeen 1000 and 1010 km
	else 
		return ( (InitialPericenter < MCC_10_PERI_MIN) || (InitialPericenter > MCC_10_PERI_MAX) );
}



void Luna9KTDU::MccAutopilotOrient()
{
	// Moon direction, close to 1 when pointing directly, controlled by pitch and roll
	VECTOR3 DirMoonGlobal, DirMoonLocal;
	GetRelativePos(hMoon, DirMoonGlobal);
	LocalRot(DirMoonGlobal, DirMoonLocal);
	DirMoonLocal /= length(DirMoonLocal);
	double YMoonDir = - DirMoonLocal.y;

	// horisontal vel direction, close to 1 when pointing correctly, controlled by yaw
	VECTOR3 VHor, DirVelLocal;
	GetHorizonAirspeedVector(VHor);
	VHor.y = 0;
	VHor /= length(VHor);
	HorizonRot(VHor, DirVelLocal);
	double ZVelDir = -DirVelLocal.z;

	// reverse Z (tangential) direction for Luna10 getting too close
	if ( (AlsType == SATELLITE) && (InitialPericenter < MCC_10_PERI_MIN) )
		ZVelDir = - ZVelDir;

	// update string
	if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "MCC Orientation: Pitch+Roll %.3f %%, Yaw %.3f %%", YMoonDir*100, ZVelDir*100);


	VECTOR3 avel;
	GetAngularVel(avel);
	avel *= DEG;
	double omMax;
	double Thrust;

		// PITCH
	
	bool bPitchUp;

	//
	if (YMoonDir > 0.9)
	{
		Thrust = 0.5;
		omMax = 0.5;
	}
	else
	{
		Thrust = 1;
		omMax = 3;
	}
	

	// positive rotation? travelling in the right positive direction with nominal speed?
	if (avel.x > 0)		
		bPitchUp = ( YMoonDir > YMoonDirLast ) && (avel.x < omMax);

	// negative rotation?
	else					// travelling in the right negative direction too fast?
		bPitchUp =  ( (YMoonDir > YMoonDirLast)  && (avel.x < -omMax) )
					||		// travelling in the wrong direction? 
					(YMoonDir < YMoonDirLast);


	// have to pitch up?
	if (bPitchUp)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, Thrust);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}

	// or pitch down?
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, Thrust);
	}


		// ROLL
	
	bool bRollRight;

	if (YMoonDir > 0.9)
	{
		Thrust = 0.3;
		omMax = 0.5;
	}
	else
	{
		Thrust = 1;
		omMax = 3;
	}


	if (avel.z > 0)		
		bRollRight = ( YMoonDir > YMoonDirLast ) && (avel.z < omMax);

	else				
		bRollRight = ( (YMoonDir > YMoonDirLast)  && (avel.z < -omMax) )
					 ||
					 (YMoonDir < YMoonDirLast);

	if (bRollRight)
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, Thrust);
	}

	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, Thrust);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	}


		// YAW
	
	bool bYawLeft;

	// NOTE: ignore Z until Y is within 90%
	if (YMoonDir < 0.9)
	{
		Thrust = 0;
		omMax = 1;
	}

	else if (ZVelDir > 0.9)
	{
		Thrust = 0.5;
		omMax = 0.5;
	}
	else
	{
		Thrust = 1;
		omMax = 3;
	}


	if (avel.y > 0)		
		bYawLeft =  (ZVelDir > ZVelDirLast) && (avel.y < omMax);

	else				
		bYawLeft = ( (ZVelDir > ZVelDirLast)  && (avel.y < -omMax) )
				   ||
				   (ZVelDir < ZVelDirLast); 

	if (bYawLeft)
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, Thrust);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	}

	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, Thrust);
	}

		 
	// we'll need it in the next timestep...
	YMoonDirLast = YMoonDir;
	ZVelDirLast = ZVelDir;

	// are we done with orienting?

	if ( (YMoonDir > 0.9999) && (ZVelDir > 0.9999) )
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);

		MccState = MCC_STEADY;
		ActivateNavmode(NAVMODE_KILLROT);
		Timer = 7;
	}
}


void Luna9KTDU::MccAutopilotWait()
{
	Timer -= Simdt;

	if (Timer < 0)
	{
		if (oapiGetFocusObject() == GetHandle())
			sprintf(oapiDebugString(), "");

		SetThrusterGroupLevel(THGROUP_MAIN, 1);
		MccState = MCC_BURN;
	}

	else if (oapiGetFocusObject() == GetHandle())
		sprintf(oapiDebugString(), "MCC Burn in %.0f s...", Timer);
}


void Luna9KTDU::MccAutopilotBurn()
{
	// check fuel
	bool bStopByFuel = GetPropellantMass(MainTank)/GetPropellantMaxMass(MainTank) < 0.90;

	// check pericenter
	ELEMENTS ElToMoon;
	ORBITPARAM OPtoMoon;
	GetElements(hMoon, ElToMoon, &OPtoMoon, 0, FRAME_EQU);
	
	bool bStopByPericenter;
	if (AlsType == LANDER)
		bStopByPericenter = OPtoMoon.PeD < MCC_9_PERI_MIN;
	
	else if (InitialPericenter < MCC_10_PERI_MIN) 
		bStopByPericenter = (OPtoMoon.PeD > MCC_10_PERI_MIN);
	
	else
		bStopByPericenter = OPtoMoon.PeD < MCC_10_PERI_MIN;

	// check pericenter var
	bool bStopByPericenterVar;
	if ( (AlsType == SATELLITE) && (InitialPericenter < MCC_10_PERI_MIN) )
		bStopByPericenterVar = PastStepPericenter > OPtoMoon.PeD;
	else
		bStopByPericenterVar = PastStepPericenter < OPtoMoon.PeD;

	// store for the next timestep
	PastStepPericenter = OPtoMoon.PeD;

	if (bStopByFuel || bStopByPericenter || bStopByPericenterVar)
	{
		MccState = MCC_NONE;
		PastStepPericenter = 0;

		DeactivateNavmode(NAVMODE_KILLROT);
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}
}


void Luna9KTDU::CalcAltBrake()
{
	// 1. Find V at the approx point of break

	// calc vel at approx break alt, 70-75 km
	VECTOR3 VHor;
	GetHorizonAirspeedVector(VHor);
	double V0 = length(VHor);

	// 1.3 is the approx moon g at the altitude influenced by common approach angles.
	// 75 k is approx breaking start,the final value won't ve very different
	double dHcurr = Alt - 75000;
	double D = V0*V0 - 2 * 1.3 * dHcurr;
	double t = (V0 - sqrt(D)) / 1.3;
	double Vbralt = V0 + t * 1.3; 

	// debug calc landing autopilot
//	sprintf(oapiDebugString(), "Vbralt: %f", Vbralt);

	// get time required to break this velocity
	double LogDM = Vbralt / ISP_MAIN;
	double DM = exp(LogDM);  
	double M0 = GetMass() - PO_MASS*2;
	double Mk = M0/DM;
	double dM = M0 - Mk;

	double dMsec = ((THRUST_MAIN + 4*THRUST_STEER + 2*THRUST_STEER_ROLL)) / ISP_MAIN;

	// find initial trajectory angle, we'll need it later
	double cosA = fabs(VHor.y)/V0;

	// find distance that we'll travel in this given time with engine on, approximate with each sec.
	double M = M0;
	double P = (THRUST_MAIN + 4*THRUST_STEER);
	double K = 4.9E+12;
	double R = 1.74E+06 + AltBrake;
	double g = K / (R * R);
	double a = P / M - g;
	double V = Vbralt;
	double L = 0; // accumulated covered distance (inclined)
	double H = 0; // accumulated covered altitude 
	int i;
	double k; // throttle

	// calc the by-the-second alt-vel program of the powered descent
	for (i = 0; i < 99; i++)
	{
		// store current step
		LApLand[i] = L;
		VApLand[i] = V;

		// calc next step

		// plan to throttle down as velocity decreeses
		if (V < 500)
			k = 0.8;
		else if (V < 1000)
			k = 0.85;
		else if (V < 2000)
			k = 0.9;
		else
			k = 1;

		H = L * cosA; // on the inclined trajectory, we cover less altitude than distance
		M -= (dMsec * k);
		R = 1.74E+06 + AltBrake - H;
		g = K / (R * R);
		a = (P * k) / M - g * cosA;
		V -= a;
		L += V - a/2;

		// stop?
		if (V < 30)
			break;
	}

	// We can now calculate breaking alttitude (plus-minus empirical adjustment)
	AltBrake = H - 700;
}



void Luna9KTDU::LandingAutopilot()
{

	// current vel
	VECTOR3 VHor;
	GetHorizonAirspeedVector(VHor);
	double V = length(VHor);

	// find i for current vel
	int I = -1;
	for (int i = 0; i < 99; i++)
		if (VApLand[i] < V)
		{
			I = i;
			break;
		}

	// check validity
	if (I == -1)
		return;

	double Vlow = VApLand[I];
	double Vhi = VApLand[I-1];
	double ratioV = (V-Vlow) / (Vhi - Vlow);

	double Hhi = LApLand[I-1];
	double Hlow = LApLand[I];

	double Hnom = Hlow + (Hhi - Hlow) * ratioV;
	
	Lap += V * Simdt;

	if (Hnom < Lap)
		SetThrusterGroupLevel(MainThGroup, Pmain += Simdt);
	else 
		SetThrusterGroupLevel(MainThGroup, Pmain -= Simdt);


	// debug landing profile
//	sprintf(oapiDebugString(), "i: %d, ratio: %f, delta: %f, revHapIntegr %f, revHnom %f, Alt %f", i, Hnom/Hap, Hnom-Hap, AltBrake-Hap+260, AltBrake-Hnom+260, Alt);
//	sprintf(oapiDebugString(), "revHnom %f, Vlow  %f, Vhi  %f", AltBrake-Hnom+260, Vlow, Vhi);
}



// only done during/after scenario load, not in-flight!
// This function run always, even if we "decide" not to attach for whatever reasons.
void Luna9KTDU::SetupPoAstro()
{
	// default to non-attached
	PoAstroMass = 0;

	// should we attach?
	if (PoAstroName[0] == 0)
		return;

	// does the object exist?
	OBJHANDLE hPo = oapiGetObjectByName(PoAstroName);
	if (!hPo)
		return;

	// get PO vessel interface
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hPo);

	// does PO have attachment? 
	if (Intf->AttachmentCount(TOPARENT) == 0)
		return;

	// get attachment handle from PO
	OBJHANDLE hToMe = Intf->GetAttachmentHandle(TOPARENT, 0);

	// Attach!
	AttachChild(hPo, hForPoAstro, hToMe);
	PoAstroMass = PO_MASS;
}



// only done during/after scenario load, not in-flight!
// This function run always, even if we "decide" not to attach for whatever reasons.
void Luna9KTDU::SetupPoRadio()
{
	// default to non-attached
	PoRadioMass = 0;

	// should we attach?
	if (PoRadioName[0] == 0)
		return;

	// does the object exist?
	OBJHANDLE hPo = oapiGetObjectByName(PoRadioName);
	if (!hPo)
		return;

	// get PO vessel interface
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hPo);

	// does PO have attachment? 
	if (Intf->AttachmentCount(TOPARENT) == 0)
		return;

	// get attachment handle from PO
	OBJHANDLE hToMe = Intf->GetAttachmentHandle(TOPARENT, 0);

	// Attach!
	AttachChild(hPo, hForPoRadio, hToMe);
	PoRadioMass = PO_MASS;

	// activate gas RCS
	SetPropellantMass(GasRCSTank, 1);
}


void Luna9KTDU::SetupAls()
{
	// default to non-attached
	AlsMass = 0;

	// should we attach?
	if (AlsName[0] == 0)
		return;

	// does the object exist?
	OBJHANDLE hAls = oapiGetObjectByName(AlsName);
	if (!hAls)
		return;

	// get child vessel interface
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hAls);

	// does child have attachment? 
	if (Intf->AttachmentCount(TOPARENT) == 0)
		return;

	// get attachment handle from child
	OBJHANDLE hToMe = Intf->GetAttachmentHandle(TOPARENT, 0);

	// Attach!
	AttachChild(hAls, hForAls, hToMe);

	// update mass
	if (AlsType == LANDER)
		AlsMass = ALS9_MASS;
	else
		AlsMass = ALS10_MASS;
}


// by J key, by autopilot or by accident :-)
void Luna9KTDU::SeparatePoAstro()
{
	// do we have someting to separate?
    if (!GetAttachmentStatus(hForPoAstro))
		return;

	// Better to randomize the kick effect
	double vDetach = 0.1; 

	// separate whatever is attached to the payload attacment, indiscriminatory
	DetachChild(hForPoAstro, vDetach);
	SendVesselMessage(PoAstroName, VM_DETACH);

	// 
	PoAstroMass = 0;

	// do the last kick 
	OBJHANDLE hPo = oapiGetObjectByName(PoAstroName);
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hPo);
	Intf->AddForce( _V(0, 10000, 0), _V(0, 0, 1) );

	// kill PO name so it won't be stored in scn file (if it will, it can reactivate the attachment)
	PoAstroName[0] = 0;

	// call universal post-separation operation
	UpdatePhysics();

	PlayVesselWave3(SoundLibID, SND_PYRO);
}


void Luna9KTDU::SeparatePoRadio()
{
	// do we have someting to separate?
    if (!GetAttachmentStatus(hForPoRadio))
		return;

	// Better to randomize the kick effect
	double vDetach = 0.1; 

	// separate whatever is attached to the payload attacment, indiscriminatory
	DetachChild(hForPoRadio, vDetach);
	SendVesselMessage(PoRadioName, VM_DETACH);

	// 
	PoRadioMass = 0;

	// do the last kick 
	OBJHANDLE hPo = oapiGetObjectByName(PoRadioName);
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hPo);
	Intf->AddForce( _V(0, -10000, 0), _V(0, 0, 1) );

	// kill PO name so it won't be stored in scn file (if it will, it can reactivate the attachment)
	PoRadioName[0] = 0;

	// disable gas RCS
	SetPropellantMass(GasRCSTank, 0);

	// call universal post-separation operation
	UpdatePhysics();

	// sounds
	PlayVesselWave3(SoundLibID, SND_PYRO);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, false);
}


void Luna9KTDU::SeparateAls()
{
	// do we have someting to separate?
    if (!GetAttachmentStatus(hForAls))
		return;

	// Better to randomize the kick effect
	double vDetach = 0.1; 

	PlayVesselWave3(SoundLibID, SND_PYRO);

	// separate whatever is attached to the payload attacment, indiscriminatory
	DetachChild(hForAls, vDetach);
	SendVesselMessage(AlsName, VM_DETACH);

	// 
	AlsMass = 0;

	// get ALS interface 
	OBJHANDLE hAls = oapiGetObjectByName(AlsName);
	VESSEL	*Intf;
	Intf = oapiGetVesselInterface(hAls);

	// for lander, give it a good kick
	if (AlsType == LANDER)
		Intf->AddForce( _V(DRand(10000, 1000), DRand(10000, 1000), DRand(50000, 5000)), _V(0, 1, 1) );

	// start ALS timer
	// We could do it in one message, VM_DETACH. 
	// But VM_DETACH is processed in VESSEL2M level, while VM_TIMER is processed in Als level.
	// rather than intercepting and re-sending the VM_DETACH message in Als, it is easier to
	// send two separate messages.
	SendVesselMessage(AlsName, VM_TIMER);

	// kill Als name so it won't be stored in scn file (if it will, it can reactivate the attachment)
	AlsName[0] = 0;

	// disable gas RCS
	SetPropellantMass(GasRCSTank, 0);

	// call universal post-separation operation
	UpdatePhysics();

	// always make sure that ALS can now accept focus
	Intf->SetEnableFocus(true);

	if (bSwitchFocus)
		oapiSetFocusObject(hAls);
}


void Luna9KTDU::UpdatePhysics()
{
	// update mass
	SetEmptyMass(EMPTY_MASS + PoAstroMass + PoRadioMass + AlsMass);
}


void Luna9KTDU::Land()
{
	// break landing probes
	SetAnimation(ExtendAnimation, 0.50);
	SetAnimation(Brake1Animation, 1);
	SetAnimation(Brake2Animation, 1); 
	
	AnimStatus = BROKEN;

	// permanently safe main engine
//	SetPropellantMass(MainTank, 0);
}


void Luna9KTDU::ToggleMainThrustChamber()
{
	//	
	if (MainThGroupCount == 6)
		EnableMainThrustChamber();
	else
		DisableMainThrustChamber();
}


void Luna9KTDU::EnableMainThrustChamber()
{
	// toggling chamber requires engine shutdown. Do we have to restart it later?
	double RestartThrust = GetThrusterGroupLevel(MainThGroup);
	
	// shutdown
	SetThrusterGroupLevel(MainThGroup, 0);

	// regroup chambers
	MainThGroupCount = 7;
	DelThrusterGroup(THGROUP_MAIN);
	MainThGroup = CreateThrusterGroup(MainEngine, MainThGroupCount, THGROUP_MAIN);

	// restart engine, if it was running
	SetThrusterGroupLevel(MainThGroup, RestartThrust);

	// show message
	ShowAnnotationMessage("Main combustion chamber activated!");

	// update engine sound
	ReplaceStockSound3(SoundLibID, "Sound\\Vessel\\mainextloud.wav", REPLACE_MAIN_THRUST);
}


void Luna9KTDU::DisableMainThrustChamber()
{
	// toggling chamber requires engine shutdown. Do we have to restart it later?
	double RestartThrust = GetThrusterGroupLevel(MainThGroup);

	// shutdown
	SetThrusterGroupLevel(MainThGroup, 0);

	// regroup chambers
	MainThGroupCount = 6;
	DelThrusterGroup(THGROUP_MAIN);
	MainThGroup = CreateThrusterGroup(MainEngine, MainThGroupCount, THGROUP_MAIN);

	// restart engine, if it was running
	SetThrusterGroupLevel(MainThGroup, RestartThrust);

	// show message
	ShowAnnotationMessage("Main combustion chamber deactivated!");

	// update engine sound
	ReplaceStockSound3(SoundLibID, "Sound\\Luna9\\mainextloud_minus.wav", REPLACE_MAIN_THRUST);
}



void Luna9KTDU::ToggleOnboardCamera(CameraDirs Dir)
{
	CameraDir = Dir;

	if (Dir == CAMERA_UP)
	{
		SetCameraOffset(_V(0.8, 0, -0.4)); // touchdown probe location
		SetCameraDefaultDirection(_V(0,0,1));
	}

	else if (Dir == CAMERA_DOWN)
	{
		SetCameraOffset(_V(0.6, 0.6, 0.35)); // flange location
		SetCameraDefaultDirection(_V(0,0,-1));
	}
}


void Luna9KTDU::InflateAirbags()
{
	if (GetAttachmentStatus(hForAls) && GetAttachmentStatus(hForPoRadio))
	{
		SendVesselMessage(AlsName, VM_INFLATE);
		PlayVesselWave3(SoundLibID, SND_INFLATE);
	}
}



void Luna9KTDU::SetTimeAccelOff()
{
	if (oapiGetTimeAcceleration() == 1)
		return;

	oapiSetTimeAcceleration(1);
}


// this CAN be called at any timestep, so don't bloat it!
void Luna9KTDU::LimitTimeAccel()
{
	// ignore if autopilot is off
	if (!bAutopilot)
		return;

	if (oapiGetTimeAcceleration() > MaxTimeAccel)
		oapiSetTimeAcceleration(MaxTimeAccel);
}





	// ===== ANIMATIONS =====


void Luna9KTDU::DefineAnimations()
{
	const double Z_PROBE = -0.64;

	static UINT grProbes[2] = {55,56}; 
	MGROUP_SCALE *Extend;
	Extend = new MGROUP_SCALE(0, grProbes, 2, _V( 0, 0, Z_PROBE), _V(1, 1, 80.));
	ExtendAnimation = CreateAnimation (0);
	AddAnimationComponent (ExtendAnimation, 0, 1, Extend);

	static UINT grProbe1[1] = {55}; 
	MGROUP_ROTATE *Brake1;
	Brake1 = new MGROUP_ROTATE(0, grProbe1, 1, _V( -0.61, 0, Z_PROBE), _V(cos(45*RAD), cos(45*RAD), 0), (float)(60*PI/180));
	Brake1Animation = CreateAnimation (0);
	AddAnimationComponent (Brake1Animation, 0, 1, Brake1);

	static UINT grProbe2[1] = {56}; 
	MGROUP_ROTATE *Brake2;
	Brake2 = new MGROUP_ROTATE(0, grProbe2, 1, _V( 0, 0, Z_PROBE), _V(1, 0, 0), -(float)(41*PI/180));
	Brake2Animation = CreateAnimation (0);
	AddAnimationComponent (Brake2Animation, 0, 1, Brake2);
}


void Luna9KTDU::Animate()
{
	double dPercent = 1 * Simdt;

	PlayVesselWave3(SoundLibID, SND_PROBE, LOOP);

	if (AnimationPercent < 1.0)
		AnimationPercent = min (1.0, AnimationPercent + dPercent);
	else
		AnimationEnd();

	SetAnimation (ExtendAnimation, AnimationPercent);
}


void Luna9KTDU::AnimationStart()
{
	// already animating something? ignore new animation request until done.
	if (bAnimate)
		return;

	// already extended or broken?
	if (AnimStatus != INITIAL)
		return;

	// lander mission?
	if (AlsType == SATELLITE)
		return;
	
	// finally, if we got there - allow animation!
	bAnimate = true;

	PlayVesselWave3(SoundLibID, SND_PYRO);
}


void Luna9KTDU::AnimationEnd()
{
	bAnimate = false;
	AnimStatus = EXTENDED;
	StopVesselWave3(SoundLibID, SND_PROBE);
}







void Luna9KTDU::LocalRot(const VECTOR3 vecGlob, VECTOR3 &vecLoc)
{
	VESSELSTATUS2 vs;
	double dCosZ, dSinZ, dCosY, dSinY, dCosX, dSinX;
	vs.version = 2;
	vs.flag = 0;
	GetStatusEx((void*) &vs);

	dCosZ = cos(vs.arot.z);
	dSinZ = sin(vs.arot.z);
	dCosY = cos(vs.arot.y);
	dSinY = sin(vs.arot.y);
	dCosX = cos(vs.arot.x);
	dSinX = sin(vs.arot.x);

	vecLoc.x = vecGlob.x*dCosZ*dCosY + vecGlob.y*(-dSinZ*dCosX+dCosZ*dSinY*dSinX) + vecGlob.z*(dSinZ*dSinX+dCosZ*dSinY*dCosX);
	vecLoc.y = vecGlob.x*dSinZ*dCosY + vecGlob.y*(dCosZ*dCosX+dSinZ*dSinY*dSinX) + vecGlob.z*(-dCosZ*dSinX+dSinZ*dSinY*dCosX);
	vecLoc.z = vecGlob.x*-dSinY + vecGlob.y*dCosY*dSinX + vecGlob.z*dCosY*dCosX;
}



// =============== OUTPUT

void Luna9KTDU::ShowAnnotationMessage(char *Message)
{
	if (oapiGetFocusObject() != GetHandle())
		return;

	oapiAnnotationSetText (MessagePort, Message);
	msgStartTime = oapiGetSysTime();
	bMessageDisplayed = true;
}


// permanent strings are set in Do components of Check-Do pairs, but are moved to a separate
// function to be also called from FinishConfiguration on scenario restore
void Luna9KTDU::ShowPermanentString(chk_do_types ChkDo, bool bIgnoreFocus)
{
	if (!bIgnoreFocus)
		if (oapiGetFocusObject() != GetHandle())
			return;

	if (ChkDo == CDT_MCCALT)
		sprintf(oapiDebugString(), "Coasting to %.0f Mm distance...", ALT_MCC_NOMINAL/1000000);

	else if ( (ChkDo == CDT_TIME1000) || (ChkDo == CDT_TIME100) || (ChkDo == CDT_TURN))
		sprintf(oapiDebugString(), "Coasting to %.0f km distance...", AltTurn/1000);

	else if (ChkDo == CDT_IGNALT)
		sprintf(oapiDebugString(), "Coasting to %.0f km ignition altitude...", AltBrake/1000);

}




void Luna9KTDU::DebugSomething()
{
//	char tempStr[256];
//	sprintf(tempStr, "MaxTimeAccel: %f", MaxTimeAccel);
//	ShowAnnotationMessage(tempStr);

//	sprintf(oapiDebugString(), "%d", GroundContact());
}


		// SOUND
void Luna9KTDU::clbkPostCreation (void)
{
	SoundLibID = ConnectToOrbiterSoundDLL3(GetHandle());

	SoundOptionOnOff3(SoundLibID, PLAYMAINTHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYLANDINGANDGROUNDSOUND, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAIRSPEED, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAMBIANCEWHENLANDED, true);

	SoundOptionOnOff3(SoundLibID, PLAYCOUNTDOWNWHENTAKEOFF, false);
	SoundOptionOnOff3(SoundLibID, PLAYWHENATTITUDEMODECHANGE, false);
	SoundOptionOnOff3(SoundLibID, PLAYGPWS, false);
	SoundOptionOnOff3(SoundLibID, PLAYHOVERTHRUST, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKINGSOUND, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADARBIP, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKLANDCLEARANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINAIRCONDITIONING, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINRANDOMAMBIANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADIOATC, false);
	SoundOptionOnOff3(SoundLibID, DISPLAYTIMER, false);

	RequestLoadVesselWave3(SoundLibID, SND_PYRO, "Sound\\Luna9\\Pyro.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_INFLATE, "Sound\\Luna9\\Inflate.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_PROBE, "Sound\\Luna9\\Probe.wav", BOTHVIEW_FADED_FAR);

	ReplaceStockSound3(SoundLibID, "Sound\\Luna9\\Gas.wav", REPLACE_RCS_THRUST_SUSTAIN);
}


// ==============================================================
// API interface
// ==============================================================

DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	srand((unsigned)time( NULL ));
	return new Luna9KTDU (hvessel, flightmodel);
}

DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) 
		delete (Luna9KTDU*)vessel;
}



